from prompt_toolkit.completion import WordCompleter
from constants.assistant_info import NOTEBOOK_COMMANDS, ADDRESSBOOK_COMMANDS

notebook_commands = [c["command"] for c in NOTEBOOK_COMMANDS]
addressbook_commands = [c["command"] for c in ADDRESSBOOK_COMMANDS]

COMBINED_BOT_COMMANDS = list(set(notebook_commands + addressbook_commands))

COMMAND_CORRECTIONS = {
    # "hello"
    "hlelo": "hello",
    "hleloo": "hello",
    "helol": "hello",
    "helo": "hello",
    "ello": "hello",
    "hllo": "hello",
    # "info"
    "nifo": "info",
    "ifno": "info",
    "inof": "info",
    "infor": "info",
    "inffo": "info",
    "infro": "info",
    "ino": "info",
    # "info-addressbook"
    "info addressbook": "info-addressbook",
    "info-adressbook": "info-addressbook",
    "info-adressbok": "info-addressbook",
    "info-adresbook": "info-addressbook",
    "info-addresbook": "info-addressbook",
    "info-addressbokk": "info-addressbook",
    # "add-contact"
    "add contact": "add-contact",
    "ad-contact": "add-contact",
    "add-contct": "add-contact",
    "add-cnotact": "add-contact",
    "add-cotnact": "add-contact",
    "add-contatc": "add-contact",
    # "delete-contact"
    "delete contact": "delete-contact",
    "delte-contact": "delete-contact",
    "delee-contact": "delete-contact",
    "deltee-contact": "delete-contact",
    "dellete-contact": "delete-contact",
    "deelte-contact": "delete-contact",
    # "phone"
    "phne": "phone",
    "phoe": "phone",
    "pone": "phone",
    "phon": "phone",
    "phoen": "phone",
    "phnone": "phone",
    # "add-phone"
    "add phone": "add-phone",
    "ad-phone": "add-phone",
    "add-phne": "add-phone",
    "add-phoen": "add-phone",
    "add-phnone": "add-phone",
    "add-ponhe": "add-phone",
    # "add-email"
    "add email": "add-email",
    "ad-email": "add-email",
    "add-emal": "add-email",
    "add-emial": "add-email",
    "add-emai": "add-email",
    "add-emaiil": "add-email",
    # "add-address"
    "add address": "add-address",
    "ad-address": "add-address",
    "add-adress": "add-address",
    "add-adres": "add-address",
    "add-adrress": "add-address",
    "add-addess": "add-address",
    # "search-contact"
    "search contact": "search-contact",
    "serch-contact": "search-contact",
    "seach-contact": "search-contact",
    "sarch-contact": "search-contact",
    "searh-contact": "search-contact",
    "searc-contact": "search-contact",
    # "add-birthday"
    "add birthday": "add-birthday",
    "ad-birthday": "add-birthday",
    "add-birtday": "add-birthday",
    "add-birthay": "add-birthday",
    "add-birhday": "add-birthday",
    # "show-birthday"
    "show birthday": "show-birthday",
    "shwo-birthday": "show-birthday",
    "show-birtday": "show-birthday",
    "show-birthay": "show-birthday",
    "show-birhday": "show-birthday",
    # "birthdays"
    "birtday": "birthdays",
    "birtdays": "birthdays",
    "brithdays": "birthdays",
    "birthdys": "birthdays",
    "brthdays": "birthdays",
    "birthdayss": "birthdays",
    # "all"
    "all contacts": "all-contacts",
    "all-contcts": "all-contacts",
    "al-cntacts": "all-contacts",
    "all-contact": "all-contacts",
    "alll-contacts": "all-contacts",
    # "info-notebook"
    "info notebook": "info-notebook",
    "info-notebokk": "info-notebook",
    "info-notebok": "info-notebook",
    "info-notbok": "info-notebook",
    "info-noteboook": "info-notebook",
    # "add-note"
    "add note": "add-note",
    "ad-note": "add-note",
    "add-not": "add-note",
    "add-noet": "add-note",
    "add-nte": "add-note",
    "add-nnote": "add-note",
    # "edit-note"
    "edit note": "edit-note",
    "edit-noet": "edit-note",
    "edit-not": "edit-note",
    "eddit-note": "edit-note",
    "edit-nte": "edit-note",
    "edit-ntoe": "edit-note",
    # "delete-note"
    "delte-note": "delete-note",
    "delee-note": "delete-note",
    "deltee-note": "delete-note",
    "dellete-note": "delete-note",
    "deelte-note": "delete-note",
    # "search-note"
    "search note": "search-note",
    "serch-note": "search-note",
    "seach-note": "search-note",
    "sarch-note": "search-note",
    "searh-note": "search-note",
    "searc-note": "search-note",
    # "all-notes"
    "all notes": "all-notes",
    "all-noets": "all-notes",
    "al-notes": "all-notes",
    "all-nots": "all-notes",
    "alll-notes": "all-notes",
    # "close"
    "clsoe": "close",
    "cose": "close",
    "clsose": "close",
    "cloose": "close",
    "cloe": "close",
    # "exit"
    "exi": "exit",
    "exitt": "exit",
    "exti": "exit",
    "eixt": "exit",
    "exiit": "exit",
}

COMPLETER = WordCompleter(COMBINED_BOT_COMMANDS, ignore_case=True, sentence=True)



from classes.field import Field


class Name(Field):
    """
    A class for storing a contact's name.
    Inherits from Field.
    """

    def __init__(self, name):
        super().__init__(name)
        self.value = name



import logging

import colorlog


def setup_logging():
    """
    Set up colored logging for the application.
    """
    handler = colorlog.StreamHandler()
    handler.setFormatter(
        colorlog.ColoredFormatter(
            "%(log_color)s%(asctime)s - %(levelname)s - %(message)s",
            datefmt=None,
            reset=True,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "bold_red",
            },
        )
    )

    logger = colorlog.getLogger()
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    return logger



"""birthday handlers"""

from helpers.assistant_info import table_show

from classes import AddressBook

from .decorators import input_error


@input_error
def add_birthday_to_contact(args, book: AddressBook):
    """Adds a birthday to the contact.

    Args:
        args (list): contains name and birthday
        book (class): contact list

    Returns:
        str: message that the birthday has been added to contact
    """
    name, date, *_ = args
    record = book.find(name)
    if record:
        record.add_birthday(date)
        return "Birthday added"
    else:
        return f"There is no contact {name}"


@input_error
def show_birthday(args, book: AddressBook):
    """Shows a birthday of the contact.

    Args:
        args (list): contains name
        book (class): contact list

    Returns:
        str: message that shows contact name and birthday
    """
    name, *_ = args
    record = book.find(name)
    if record:
        if not record.birthday:
            return f"There is no birthday for contact {name}"
        return f"{name} birthday: {record.birthday}"

    else:
        return f"There is no contact {name}"


@input_error
def show_upcoming_birthdays(book: AddressBook):
    """Shows all contacts with congratulations dates for the next X days.

    Args:
        book (AddressBook): contact list.

    Returns:
        str: message with a list of contacts with congratulations dates.
    """
    try:
        days = int(
            input(
                "Enter the number of days from today to check for upcoming birthdays: "
            ).strip()
        )
    except ValueError:
        return "Invalid input! Please enter a valid number of days."

    upcoming_birthdays = book.get_upcoming_birthdays(days)

    if len(upcoming_birthdays) == 0:
        return "No contacts that need to be congratulated within the specified period."

    headers = ["Name", "Congratulation date"]
    table_data = [
        [key["name"], key["congratulation_date"]] for key in upcoming_birthdays
    ]
    return table_show(headers, table_data)



from classes import NotesBook, Note
from helpers.assistant_info import table_show

headers = ["Note Name", "Text", "Tags"]
header = ["Tags"]

def handle_errors(func):
    """
    Decorator to handle errors and provide user-friendly messages.

    Args:
        func (function): The function to wrap with error handling.

    Returns:
        function: The wrapped function with error handling.
    """

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyError as e:
            return f"Error: {str(e)}"
        except ValueError as e:
            return f"Invalid input: {str(e)}"
        except Exception as e:
            return f"An unexpected error occurred: {str(e)}"

    return wrapper


@handle_errors
def add_note(notebook: NotesBook):
    """
    Adds a new note to the notes list.

    Args:
        notebook (NotesBook): The notes list.

    Returns:
        str: Message indicating that the note has been added or an error occurred.
    """
    name = input("Enter note name: ").strip()
    if name:
        text = input("Enter note text: ")
        if text:
            note = notebook.find(name)
            if note is None:
                note = Note(name, text)
                notebook.add_note(note)
                return "Note added."
            else:
                return f"Note with name '{name}' already exists."
        else:
            return "Text cannot be empty."
    else:
        return "Name cannot be empty."


@handle_errors
def show_all_notes(notebook: NotesBook):
    """
    Shows all notes in the list.

    Args:
        notebook (NotesBook): The notes list.

    Returns:
        str: Formatted string with all notes or an error message.
    """
    if not notebook.data:
        return "No notes found."

    rows = [(note.name.value, note.text.value, ' '.join(note.tags)) for note in notebook.data.values()]
    return table_show(headers, rows)


@handle_errors
def find_note(notebook: NotesBook):
    """Finds and returns notes by searching for a keyword in their
    names from the notebook.
    Handles user input directly.
    Args:
        notebook (NotesBook): The notebook to search in.
    Returns:
        str: The notes that match the search, or a message if no matches are found."""

    while True:
        note_name = input("Enter the note name or keyword: ").strip().lower()
        if note_name:
            matching_notes = [
                note
                for note in notebook.data.values()
                if note_name in str(note.name).lower()
            ]
            if matching_notes:
                rows = [(str(note.name), str(note.text), " ".join(note.tags)) for note in matching_notes]
                return table_show(headers, rows)
            else:
                return "No notes found matching that keyword."
        if note_name == "no":
            break


@handle_errors
def edit_note(notebook: NotesBook):
    """
    Edits the text of an existing note in the notebook.

    Prompts the user to enter the name of the note they wish to edit.
    If the note is found, it asks for the new text and updates the note.
    If the note is not found, or if the new text is empty, an appropriate
    message is returned.

    Args:
        notebook (NotesBook): The notebook instance containing notes.

    Returns:
        str: The updated note if the operation is successful, or an error message
            if the note is not found or if the new text is empty.
    """
    name = input("Enter note name: ").strip()
    note = notebook.find(name)
    if note:
        while True:
            usr_chose = input("What do you want to change: (name/text): ").strip()
            if usr_chose == "name":
                while True:
                    new_name = input("Type new name or back to go back: ").strip()
                    if new_name == "back":
                        break
                    elif new_name:
                        note.edit_name(new_name)
                        notebook.add_note(note)
                        notebook.delete(name)
                        return "Name changed."
                    else:
                        print("Type the name, should not be empty.")

            elif usr_chose == "text":
                while True:
                    text = input("Type new text or back to go back: ").strip()
                    if text == "back":
                        break
                    elif text:
                        note.edit_note(text)
                        return "Text changed."
                    else:
                        print("Type the name, should not be empty.")

            elif usr_chose == "exit":
                return "Editing completed."

            else:
                print("Command not found. If you don't want to edit note type | exit |")
    else:
        return f"Note {name} haven't found!"


def remove_note(note_name: str, notebook: NotesBook):
    """
    Removes a note from the notebook.
    Prompts the user to enter the name of the note they wish to remove.
    If the note is found, it is removed from the notebook.
    If the note is not found, an appropriate message is returned.

    Args:
        notebook (NotesBook): The notebook instance containing notes.

    Returns:
        str: message
    """
    if note_name:
        note = notebook.find(note_name)
        if note:
            notebook.delete(note_name)
            return f"Note {note_name} has been deleted."

        if note_name == "all":
            notebook.data.clear()
            return "All notes have been deleted."
        else:
            return "Note not found."
    else:
        return "Please enter: remove-note <name> or remove-note <all>."


@handle_errors
def add_tag(note, tag, notebook: NotesBook):
    """
    Adds a tag to a note in the notebook.

    Args:
        note (str): The name of the note to add the tag to.
        tag (str): The tag to add to the note.
        notebook (NotesBook): The notebook to add the tag to.

    Returns:
        None
    """
    return notebook.add_tag(note, tag)

@handle_errors
def add_tags(args, notebook: NotesBook):
    """
    Adds multiple tags to a note in the notebook.

    Args:
        args (tuple): A tuple containing the note name and a list of tags to add.
        notebook (NotesBook): The notebook to add the tags to.

    Returns:
        None
    """
    note_name, tags = args 
    return notebook.add_tags(note_name, tags)

@handle_errors
def edit_tag(note, old_tag, new_tag, notebook: NotesBook):
    """
    Edits a tag on a note in the notebook.

    Args:
        note (str): The name of the note to edit the tag on.
        old_tag (str): The old tag to replace.
        new_tag (str): The new tag to replace the old tag with.
        notebook (NotesBook): The notebook to edit the tag in.

    Returns:
        None
    """
    return notebook.edit_tag(note, old_tag, new_tag)

@handle_errors
def remove_tag(note, tag, notebook: NotesBook):
    """
    Removes a tag from a note in the notebook.

    Args:
        note (str): The name of the note to remove the tag from.
        tag (str): The tag to remove from the note.
        notebook (NotesBook): The notebook to remove the tag from.

    Returns:
        None
    """
    return notebook.remove_tag(note, tag)

@handle_errors
def remove_tags(args, notebook: NotesBook):
    """
    Removes multiple tags from a note in the notebook.

    Args:
        args (tuple): A tuple containing the note name and a list of tags to remove.
        notebook (NotesBook): The notebook to remove the tags from.

    Returns:
        None
    """
    note_name, tags = args 
    return notebook.remove_tags(note_name, tags)

@handle_errors
def note_tags(note, notebook: NotesBook):
    """
    Retrieves the tags associated with a note in the notebook.

    Args:
        note (str): The name of the note to retrieve the tags for.
        notebook (NotesBook): The notebook to retrieve the tags from.

    Returns:
        list: A list of tags associated with the note.
    """
    tags = notebook.find_note_tags(note)
    if len(tags) > 0:
        rows = [[tag] for tag in tags]
        return table_show(header, rows)
    else:
        return 'No tags found for this note'

@handle_errors
def all_tags(notebook: NotesBook):
    """
    Retrieves all tags in the notebook.

    Args:
        notebook (NotesBook): The notebook to retrieve the tags from.

    Returns:
        list: A list of all tags in the notebook.
    """
    tags = notebook.all_tags()
    if len(tags) > 0:
        rows = [[tag] for tag in tags]
        return table_show(header, rows)
    else:
        return 'No tags at all' 

@handle_errors
def sort_by_tag(order:str, notebook: NotesBook):
    """
    Sorts the notes in the notebook by tag in the specified order.

    Args:
        order (str): The order in which to sort the notes (e.g. "asc" or "desc")
        notebook (NotesBook): The notebook to sort

    Returns:
        str: A string representation of the sorted notebook
    """
    sorted_notes = notebook.sort_by_tag(order)
    if sorted_notes:
        rows = [(note.name.value, note.text.value, ' '.join(note.tags)) for note in sorted_notes]
        return table_show(headers, rows)
    else:
        return 'No notes found'


"""All classes of the bot are collected here"""

from classes.address_book import AddressBook
from classes.record import Record
from classes.notes_book import NotesBook
from classes.note import Note

__all__ = ["AddressBook", "Record", "NotesBook", "Note"]



"""The function parses the string entered by the user into a command and its arguments.
    """

from .decorators import input_error


@input_error
def parse_input(user_input: str):
    """The function parses the string entered by the user
    into a command and its arguments.

    Args:
        user_input (str): accepts the string entered by the user in the console

    Returns:
        cmd (str), *args (list): the name of the command and
        the list of arguments in lower case
    """

    cmd, *args = user_input.lower().split()
    cmd = cmd.strip()

    return cmd, *args



"""Functions for validation inputs

    Returns:
        bool: returns result of expretions True or False
    """

from datetime import datetime
from classes.notes_book import NotesBook
import re


def input_name_validation(user_input):
    """validation input data for add name

    Args:
        user_input (str): user input in string format

    Returns:
        str: returns user input if passed validation
    """
    if len(user_input) > 0:
        return user_input


def input_number_validation(user_input):
    """validation input data for add number

    Args:
        user_input (str): user input in string format

    Returns:
        str: returns user input if passed validation
    """

    pattern = r"(^[+0-9]{1,3})*([0-9]{10,11}$)"
    match = re.search(pattern, user_input, re.IGNORECASE)
    if match:
        return user_input


def input_email_validation(user_input):
    """validation input data for add email

    Args:
        user_input (str): user input in string format

    Returns:
        str: returns user input if passed validation
    """
    pattern = r"\w+@\w+\.\w+"
    match = re.search(pattern, user_input, re.IGNORECASE)
    if match:
        return user_input


def input_address_validation(user_input):
    """validation input data for add address

    Args:
        user_input (str): user input in string format

    Returns:
        str: returns user input if passed validation
    """
    if len(user_input) > 0:
        return user_input


def input_birthday_validation(user_input):
    """validation input data for add bitrhday

    Args:
        user_input (str): user input in string format

    Returns:
        bool: returns result of expretions True or False
    """
    res = True
    try:
        res = bool(datetime.strptime(user_input, "%d.%m.%Y"))
        return res
    except ValueError:
        res = False
        return res

def input_note_validation(note_name: str, book: NotesBook):
    if book.find(note_name) is not None:
        return True
    else: 
        return False

def input_tag_validation(tag: str):
    """
    Validates a tag input by checking if it matches the expected format.

    Args:
        tag (str): The tag to validate.

    Returns:
        bool: True if the tag is valid, False otherwise.
    """
    return re.match(r"#\w+", tag)


from prompt_toolkit.key_binding import KeyBindings
from constants import COMMAND_CORRECTIONS


bindings = KeyBindings()


@bindings.add("enter")
def _(event):
    buffer = event.current_buffer
    input_text = buffer.text.strip()

    if input_text in COMMAND_CORRECTIONS:
        buffer.text = COMMAND_CORRECTIONS[input_text]

    if buffer.complete_state and buffer.complete_state.current_completion:
        buffer.apply_completion(buffer.complete_state.current_completion)
    else:
        event.app.exit(result=buffer.text)



"""
This module provides helper functions for working with the assistant,
including retrieving information, setting up logging, and handling data
serialization.

It includes the following functions and objects:

- `assistant_info`: Returns information about the assistant, such as name,
version, and description.
- `setup_logging`: Configures logging in the application. Allows flexible
configuration of log levels, output format, and log storage.
- `save_data`: Serializes and saves data to a file using pickle.
- `load_data`: Loads and deserializes data from a file saved with pickle.

These functions are available for import through the `__all__` declaration.
"""

from helpers.assistant_info import table_show
from helpers.logging_config import setup_logging
from helpers.pickle_utils import load_data, save_data
from helpers.notes_pickle_utils import load_notes, save_notes
from helpers.autocomplete_config import bindings
from helpers.startup_shutdown import welcome, good_bye, display_ascii_welcome_art

__all__ = [
    "setup_logging",
    "save_data",
    "load_data",
    "table_show",
    "save_notes",
    "load_notes",
    "bindings",
    "welcome",
    "good_bye",
    "display_ascii_welcome_art",
]



"""This module provides constants."""

from constants.assistant_info import (
    ADDRESSBOOK_INFO_TABLE_DATA,
    ADDRESSBOOK_INFO_TABLE_HEADERS,
    NOTEBOOK_INFO_TABLE_DATA,
    NOTEBOOK_INFO_TABLE_HEADERS,
)

from constants.auto_complete_cl import COMPLETER, COMMAND_CORRECTIONS

__all__ = [
    "ADDRESSBOOK_INFO_TABLE_DATA",
    "ADDRESSBOOK_INFO_TABLE_HEADERS",
    "NOTEBOOK_INFO_TABLE_DATA",
    "NOTEBOOK_INFO_TABLE_HEADERS",
    "COMPLETER",
    "COMMAND_CORRECTIONS",
]



"""A base class for fields in a record."""


class Field:
    """
    A base class for fields in a record.

    Attributes:
        * value (str): The value of the field.
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return str(self.value)



"""class NotesBook"""

from collections import UserDict
from classes.note import Note
import re


class NotesBook(UserDict):
    """
    A class to manage a collection of notes.

    Inherits from UserDict to store and manage Note instances.

    Attributes:
        data (dict): A dictionary to store notes using their name as the key.
    """

    def __init__(self):
        """
        Initializes a NotesBook instance.
        """
        super().__init__()

    def __str__(self):
        """
        Returns a string representation of all notes in the NotesBook.

        Returns:
            str: A formatted string with each note's details.
        """
        lines = [str(note) for note in self.data.values()]
        return "\n".join(lines)

    def add_note(self, note):
        """
        Adds a note to the NotesBook.

        Args:
            note (Note): The note to add.

        Raises:
            KeyError: If a note with the same name already exists.
        """
        if note.name.value in self.data:
            raise KeyError(f"Note with name '{note.name.value}' already exists.")
        self.data[note.name.value] = note

    def find(self, name) -> Note:
        """
        Finds a note by its name.

        Args:
            name (str): The name of the note to find.

        Returns:
            Note or None: The note if found, or None if not found.
        """
        return self.data.get(name, None)

    def delete(self, note_name):
        """
        Deletes a note by name.

        Args:
            * name(str) - The name of the note to be deleted.
        """
        if note_name not in self.data:
            raise KeyError(f"Note with name '{note_name}' not found.")
        del self.data[note_name]
    
    def add_tag(self, note_name, tag) -> str:
        """
        Adds a tag to a note with the given name.

        Args:
            note_name (str): The name of the note to add the tag to.
            tag (str): The tag to add to the note.

        Returns:
            str: The result of adding the tag, either a success message or an error message
                indicating that the note was not found.

        Notes:
            If the note is found, this method delegates to the note's `add_tag` method.
            If the note is not found, returns an error message indicating that the note was not found.
        """
        note = self.find(note_name)
        if note:
            return note.add_tag(tag)
        else:
            return f"Note with this name: {note_name} not found"
        
    def add_tags(self, note_name, tags: str) -> str:
        """
        Adds tags to a note with the given name.

        Args:
            note_name (str): The name of the note to add tags to.
            tags (str): A string of tags to add, separated by commas.

        Returns:
            str: A success message if the tags are added, or an error message if the note is not found.
        """
        note = self.find(note_name)
        if note:
             tags = self.__split_tags(tags)
             for tag in tags:
                note.add_tag(tag)
             return "Tags added succesfully"
        else:
             return f"Note with this name: {note_name} not found"

    def edit_tag(self, note_name, old_tag, new_tag) -> str:
        """
        Edits a tag in a note with the given name.

        Args:
            note_name (str): The name of the note to edit the tag in.
            old_tag (str): The old tag to replace.
            new_tag (str): The new tag to replace with.

        Returns:
            str: A success message if the tag is edited, or an error message if the note is not found.
        """
        note = self.find(note_name)
        if note:
            return note.edit_tag(old_tag, new_tag)
        else:
            return f"Note with this name: {note_name} not found"        

        
    def remove_tag(self, note_name, tag) -> str:
        """
        Removes a tag from a note with the given name.

        Args:
            note_name (str): The name of the note to remove the tag from.
            tag (str): The tag to remove.

        Returns:
            str: A success message if the tag is removed, or an error message if the note is not found.
        """
        note = self.find(note_name)
        if note:
            return note.remove_tag(tag)
        else:
            return f"Note with this name: {note_name} not found"
        
    def remove_tags(self, note_name, tags: str) -> str:
        """
        Removes multiple tags from a note with the given name.

        Args:
            note_name (str): The name of the note to remove the tags from.
            tags (str): A string of tags to remove, separated by commas.

        Returns:
            str: A success message if the tags are removed, or an error message if the note is not found.
        """
        note = self.find(note_name)
        if note:
             tags = self.__split_tags(tags)
             for tag in tags:
                    note.remove_tag(tag)
             return "Tags removed succesfully"
        else:
             return f"Note with this name: {note_name} not found"     
    
    def find_tag(self, note_name, tag) -> str:
        """
        Finds a tag in a note with the given name.

        Args:
            note_name (str): The name of the note to find the tag in.
            tag (str): The tag to find.

        Returns:
            str: A success message if the tag is found, or an error message if the note is not found.
        """
        note = self.find(note_name)
        if note:
            return note.find_tag(tag)
        else:
            return f"Note with this name: {note_name} not found"
        
    def find_note_tags(self, note_name) -> str:
        """
        Finds all tags in a note with the given name.

        Args:
            note_name (str): The name of the note to find the tags in.

        Returns:
            str: A string of all tags in the note, or an error message if the note is not found.
        """
        note = self.find(note_name)
        if note:
            return note.tags
        else:
            return f"Note with this name: {note_name} not found"
        
    def all_tags(self) -> set:
        """
        Finds all unique tags across all notes.

        Returns:
            set: A set of all unique tags.
        """
        tags = set()
        if len(self.data) > 0:
            for note in self.data.values():
                tags.update(note.tags)
        return tags
    
    def sort_by_tag(self, order: str = 'asc'):
        """
        Sorts notes in ascending or descending order by tags.

        Args:
            order (str): The order to sort in ('asc' or 'desc'). Default is 'asc'.

        Returns:
            list: A list of notes in the specified order.
        """
        if order == 'asc':
            return sorted(self.data.values(), key=lambda x: list(x.tags)[0])
        elif order == 'desc':
            return sorted(self.data.values(), key=lambda x: list(x.tags)[0], reverse=True)
        else:
            raise ValueError("Invalid order parameter. Must be 'asc' or 'desc'.")
            
    def __split_tags(self, tags: str) -> list:
        """
        Splits a string of tags into individual tags.' or 'desc'.")

        Args:
            tags (str): A string of tags, separated by something.

        Returns:
            list: A list of individual tags.
        """
        return re.findall(r"#\w+", tags)



"""Phone class"""

from classes.field import Field

from handlers.validations import (
    input_number_validation,
)


class Phone(Field):
    """Phone class"""

    def __init__(self, number):
        super().__init__(number)
        self.value = self.validate_number(number)

    def validate_number(self, number):
        """validates a phone number

        Args:
            number (int): phone number for validation

        Raises:
            ValueError: The phone number must contain 10 digits
            ValueError: The phone number must contain only numbers

        Returns:
            int: validated number
        """
        if input_number_validation(number) is False:
            raise ValueError(
                "The phone number must contain only "
                "numbers and lenth must contain 10 digits"
            )
        return number



"""This module provides handler functions for parsing input, managing contacts,
and handling birthdays."""

from handlers.birthday_handlers import (
    add_birthday_to_contact,
    show_birthday,
    show_upcoming_birthdays,
)

from handlers.contacts_handlers import (
    add_contact,
    change_contact,
    delete_contact,
    show_all,
    show_phone,
    search_contact,
    add_email_to_contact,
    add_phone_to_contact,
    add_address_to_contact,
)

from handlers.notes_handlers import (
    add_note,
    show_all_notes,
    edit_note,
    remove_note,
    find_note,
    add_tag,
    add_tags,
    remove_tag,
    remove_tags,
    edit_tag,
    all_tags,
    note_tags,
    sort_by_tag
)

from handlers.parse_input import parse_input
from handlers.input_handlers.add_contact_input import (
    add_contact_input
)

from handlers.input_handlers.edit_contact_input import (
    edit_contact_input
)


from handlers.input_handlers.tag_input import (
    add_tag_input, 
    add_tags_input,
    remove_tag_input,
    remove_tags_input,
    edit_tag_input,
    search_note_tags,
    sort_by_tag_input
)

from handlers.validations import (
    input_name_validation,
    input_number_validation,
    input_email_validation,
    input_address_validation,
    input_birthday_validation,
    input_tag_validation,
)

__all__ = [
    "parse_input",
    "add_contact",
    "change_contact",
    "show_all",
    "show_phone",
    "delete_contact",
    "add_birthday_to_contact",
    "show_birthday",
    "show_upcoming_birthdays",
    "add_note",
    "show_all_notes",
    "search_contact",
    "add_email_to_contact",
    "add_phone_to_contact",
    "add_address_to_contact",
    "add_contact_input",
    "input_name_validation",
    "input_number_validation",
    "input_email_validation",
    "input_address_validation",
    "input_birthday_validation",
    "edit_note",
    "remove_note",
    "find_note",
    "edit_contact_input",
    "add_tag",
    "add_tags",
    "remove_tag",
    "remove_tags",
    "edit_tag",
    "all_tags",
    "note_tags",
    "input_tag_validation",
    "add_tag_input",
    "add_tags_input",
    "remove_tag_input",
    "remove_tags_input",
    "edit_tag_input",
    "search_note_tags",
    "sort_by_tag",
    "sort_by_tag_input"
]



"""A module for working with a list of contacts:
adding, editing, outputting, deleting."""

from helpers.assistant_info import table_show

from classes import AddressBook, Record

from .decorators import empty_contact_list, input_error


@input_error
def add_contact(name, book: AddressBook):
    """Adds a new contact to the contact list.

    Args:
        args (list): contains name and phone number
        book (class): contact list

    Returns:
        str: message that the contact has been added or updated
    """
    record = book.find(name)
    if record is None:
        record = Record(name)
        book.add_record(record)
        message = "Contact added."
    return message


@empty_contact_list
@input_error
def change_contact(args, book: AddressBook):
    """Changes a contact's phone number by the name

    #     Args:
    #         args (list): contains name, old phone number and new phone number
    #         book (class): contact list

    #     Returns:
    #         str: notification that the contact has been changed, or contact not found
    #"""
    name, old_number, new_number, *_ = args
    record = book.find(name)
    if record is None:
        raise KeyError(f"No such name '{name}' was found")
    record.edit_phone(old_number, new_number)
    return "Phone changed"


@empty_contact_list
@input_error
def show_phone(args, book: AddressBook):
    """Shows a contact information

    Args:
        args (list): contains name and phone number
        book (class): contact list

    Returns:
        str: a message with information about the contact,
        or that the contact was not found
    """
    name, *_ = args
    record = book.find(name)
    if record is None:
        raise KeyError(f"Contact {name} not found.")
    return record


@empty_contact_list
def show_all(book: AddressBook):
    """Shows all contacts in the list

    Args:
        contacts (dict): contact list

    Returns:
        str: message with a list of contacts
    """
    headers = ["Address Book"]
    return table_show(headers, book.data.items())


@empty_contact_list
@input_error
def delete_contact(args, book: AddressBook):
    """Function to delete one contact or all at once

    Args:
        args (list): contains name and phone number
        book (class): contact list

    Returns:
        str: message about deleting one or all contacts
    """
    name, *_ = args

    record = book.find(name)
    if record:
        book.delete(name)
        return f"The {name} has been deleted"

    if name == "all":  # видалити всі контакти
        book.data.clear()
        return "All contacts have been deleted"

    return f"The {name} is not found"


@empty_contact_list
@input_error
def search_contact(args, book: AddressBook):
    """Search a contact in the book by name, phone or email

    Args:
        * args (list): contains name, phone number or email

    Returns:
        * record: founded contact or warning message
    """
    search_string, *_ = args
    search_string = search_string.strip()
    message = f"No contact with data '{search_string}' was found"
    record_by_name = book.find(search_string)
    if record_by_name:
        return record_by_name
    record_by_phone = book.find_by_phone(search_string)
    if record_by_phone:
        return record_by_phone
    record_by_email = book.find_email(search_string)
    if record_by_email:
        return record_by_email
    return message


@input_error
def add_email_to_contact(args, book: AddressBook):
    """Function to add email to contact

    Args:
        args (list): contains name and email
        book (class): contact list

    Returns:
        str: message email added
    """
    name, email, *_ = args
    record = book.find(name)
    if record:
        record.add_email(email)
        return "Email added"
    else:
        return f"There is no contact {name}"


@empty_contact_list
@input_error
def add_phone_to_contact(args, book: AddressBook):
    """Function to add phone to contact

    Args:
        args (list): contains name and phone number
        book (class): contact list

    Returns:
        str: message phone added
    """
    name, number, *_ = args
    record = book.find(name)
    if record:
        record.add_phone(number)
        return "Phone added"
    else:
        return f"There is no contact {name}"


@empty_contact_list
@input_error
def add_address_to_contact(args, book: AddressBook):
    """Function to add address to contact

    Args:
        args (list): contains name and address
        book (class): contact list

    Returns:
        str: message address added
    """
    name, *address = args
    if len(address) == 1:
        address_for_record = address[0]
    else:
        address_for_record = ""
        for world in address:
            address_for_record += f"{world} "
    record = book.find(name)
    if record:
        record.add_address(address_for_record)
        return "Address added"
    else:
        return f"There is no contact {name}"



"""A class for storing a contact's birthday."""

from datetime import datetime

from classes.field import Field

from handlers.validations import (
    input_birthday_validation,
)


class Birthday(Field):
    """
    A class for storing a contact's birthday.

    Inherits from Field. Validates that the birthday is in the format DD.MM.YYYY and
    converts the string representation to a datetime object.

    Method:
        __init__(self, value: str) - Initializes the birthday with a validated date.
    """

    def __init__(self, value: str):

        if input_birthday_validation(value):
            try:
                birthday = datetime.strptime(value, "%d.%m.%Y")
                super().__init__(birthday)
            except ValueError as exc:
                raise ValueError("Invalid date format. Use DD.MM.YYYY") from exc
        else:
            raise ValueError("Invalid date format. Use DD.MM.YYYY")

    def __str__(self):
        return f'{self.value.strftime("%d.%m.%Y")}'



import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from prompt_toolkit import prompt
from handlers import (
    add_note,
    delete_contact,
    parse_input,
    show_all,
    show_birthday,
    show_phone,
    show_upcoming_birthdays,
    show_all_notes,
    search_contact,
    add_contact_input,
    edit_contact_input,
    edit_note,
    remove_note,
    find_note,
)
from helpers import setup_logging
from helpers import (
    load_data,
    save_data,
    load_notes,
    save_notes,
    bindings,
    table_show,
    welcome,
    good_bye,
    display_ascii_welcome_art,
)
from constants import (
    ADDRESSBOOK_INFO_TABLE_DATA,
    ADDRESSBOOK_INFO_TABLE_HEADERS,
    NOTEBOOK_INFO_TABLE_DATA,
    NOTEBOOK_INFO_TABLE_HEADERS,
    COMPLETER,
)

logger = setup_logging()


def main():
    """The main function of the bot, manages the main cycle of command processing"""
    book = load_data()
    notes_book = load_notes()
    welcome()
    try:
        while True:
            user_input = prompt(
                "Enter a command: > ",
                completer=COMPLETER,
                complete_while_typing=True,
                key_bindings=bindings,
                multiline=True,
            )
            command, *args = parse_input(user_input)

            if command in ["close", "exit"]:
                save_data(book)
                save_notes(notes_book)
                good_bye()
                break

            if command == "hello":
                print("How can I help you?")

            elif command == "info":
                print(
                    table_show(
                        ADDRESSBOOK_INFO_TABLE_HEADERS,
                        ADDRESSBOOK_INFO_TABLE_DATA,
                        True,
                    )
                )
                print(
                    table_show(
                        NOTEBOOK_INFO_TABLE_HEADERS, NOTEBOOK_INFO_TABLE_DATA, True
                    )
                )

            elif command == "info-addressbook":
                print(
                    table_show(
                        ADDRESSBOOK_INFO_TABLE_HEADERS,
                        ADDRESSBOOK_INFO_TABLE_DATA,
                        True,
                    )
                )

            elif command == "info-notebook":
                print(
                    table_show(
                        NOTEBOOK_INFO_TABLE_HEADERS, NOTEBOOK_INFO_TABLE_DATA, True
                    )
                )

            elif command == "add-note":
                print(add_note(notes_book))

            elif command == "add-contact":
                add_contact_input(book)
                save_data(book, "addressbook.pkl", False)
                print("Contact saved!")

            elif command == "edit-contact":
                edit_contact_input(args, book)
                save_data(book, "addressbook.pkl", False)
                print("Contact updated!")

            elif command == "phone":
                print(show_phone(args, book))

            elif command == "all-contacts":
                print(show_all(book))

            elif command == "all-notes":
                print(show_all_notes(notes_book))

            elif command == "search-note":
                print(find_note(notes_book))

            elif command == "edit-note":
                print(edit_note(notes_book))

            elif command == "delete-note":
                note_name = " ".join(args).strip()
                print(remove_note(note_name, notes_book))

            elif command == "delete-contact":
                print(delete_contact(args, book))

            elif command == "show-birthday":
                print(show_birthday(args, book))

            elif command == "birthdays":
                print(show_upcoming_birthdays(book))

            elif command == "search-contact":
                print(search_contact(args, book))

            else:
                print("Invalid command.")
    except KeyboardInterrupt:
        save_data(book)
        save_notes(notes_book)
        display_ascii_welcome_art("Good bye")


if __name__ == "__main__":
    main()



"""class Note"""

from classes.field import Field

class Note:
    """
    A class representing a note with a name and text.

    Attributes:
        name (Field): The name of the note.
        text (Field): The text content of the note.
    """

    def __init__(self, name, text=""):
        """
        Initializes a Note instance with a name and optional text.

        Args:
            name (str): The name of the note.
            text (str, optional): The text content of the note. Defaults to empty string.
        """
        self.text = Field(text)
        self.name = Field(name)
        self.__tags = set()

    def __str__(self):
        """
        Returns a string representation of the note.

        Returns:
            str: A formatted string with the note's name and text.
        """
        return f"Note: {self.name.value}, Text: {self.text.value}, Tag: {self.tags}"
    
    def __repr__(self):
        return f"Note: {self.name.value}, Text: {self.text.value}, Tag: {self.tags}"
    
    def __getstate__(self):
        return {'text': self.text, 'name': self.name, 'tags': self.__tags}

    def __setstate__(self, state):
        self.text = state['text']
        self.name = state['name']
        self.__tags = state['tags']

    def edit_name(self, new_name):
        """A function for editing note name"""
        if new_name:
            self.name.value = new_name

    def edit_note(self, new_text):
        """A function for editing note text"""
        if new_text:
            self.text.value = new_text
        return f"Note: {self.name.value}, Text: {self.text.value}, Tag: {self.tags}"
    
    @property
    def tags(self) -> set:
        """A function for getting note tags"""
        return self.__tags
    
    def add_tag(self, tag: str) -> str:
        """A function for adding a tag to the note"""
        if not self.has_tag(tag):
            self.__tags.add(tag)
            return f"{tag} added successfully"
        else:
            return f"{tag} already is used for note"

        
    def remove_tag(self, tag: str):
        """A function for removing a tag from the note"""
        if self.has_tag(tag):
            self.__tags.remove(tag)
            return f"{tag} removed successfully"
        else:
            return f"{tag} not use for this note"
        
    def edit_tag(self, old_tag: str, new_tag: str):
        """A function for editing a tag in the note"""
        if self.has_tag(old_tag):
            self.__tags.remove(old_tag)
            self.__tags.add(new_tag)
            return f"{old_tag} changed successfully"
        else:
            return f"{old_tag} not use for this note"
        
    def find_tag(self, tag: str) -> str:
        """A function for finding a tag in the note"""
        if self.has_tag(tag):
            return self.__str__()
        
    def has_tag(self, tag: str) -> bool:
        """A function for checking if a tag is in the note"""
        return tag in self.__tags      



ADDRESSBOOK_COMMANDS = [
    {"command": "hello", "usage": "hello", "exmp": "hello", "desc": "Greets the user."},
    {
        "command": "info",
        "usage": "info",
        "exmp": "info",
        "desc": (
            "Provides general information about the address book "
            "and the notebook commands."
        ),
    },
    {
        "command": "info-addressbook",
        "usage": "info-addressbook",
        "exmp": "info-addressbook",
        "desc": "Displays information about the address book.",
    },
    {
        "command": "add-contact",
        "usage": "add-contact",
        "exmp": "add-contact",
        "desc": "Adds a new contact to the address book.",
    },
    {
        "command": "edit-contact",
        "usage": "edit-contact <name>",
        "exmp": "edit-contact bob",
        "desc": "Edits contact.",
    },
    {
        "command": "delete-contact",
        "usage": "delete-contact <name> OR delete all",
        "exmp": "delete-contact bob OR delete all",
        "desc": "Deletes a specific contact or all contacts from the address book.",
    },
    {
        "command": "phone",
        "usage": "phone <name>",
        "exmp": "phone bob",
        "desc": "Displays the phone number of the specified contact.",
    },
    {
        "command": "search-contact",
        "usage": "search-contact <name/phone/email>",
        "exmp": "search-contact bob",
        "desc": "Searches for a contact by name, phone number, or email address.",
    },
    {
        "command": "show-birthday",
        "usage": "show-birthday <name>",
        "exmp": "show-birthday bob",
        "desc": "Displays the birthday of the specified contact.",
    },
    {
        "command": "birthdays",
        "usage": "birthdays",
        "exmp": "birthdays",
        "desc": "Lists all upcoming birthdays in the range of given days.",
    },
    {
        "command": "all-contacts",
        "usage": "all-contacts",
        "exmp": "all-contacts",
        "desc": "Displays all contacts in the address book.",
    },
    {
        "command": "close",
        "usage": "close",
        "exmp": "close",
        "desc": "Saving Addressbook data and closes the application.",
    },
    {
        "command": "exit",
        "usage": "exit",
        "exmp": "exit",
        "desc": "Exits the address book application.",
    }
]

ADDRESSBOOK_INFO_TABLE_HEADERS = ["COMMAND", "USAGE", "EXAMPLE", "DESCRIPTION"]
ADDRESSBOOK_INFO_TABLE_DATA = [
    [c["command"], c["usage"], c["exmp"], c["desc"]] for c in ADDRESSBOOK_COMMANDS
]

NOTEBOOK_COMMANDS = [
    {"command": "hello", "usage": "hello", "exmp": "hello", "desc": "Greets the user."},
    {
        "command": "info",
        "usage": "info",
        "exmp": "info",
        "desc": (
            "Provides general information about the address book "
            "and the notebook commands."
        ),
    },
    {
        "command": "info-notebook",
        "usage": "info-notebook",
        "exmp": "info-notebook",
        "desc": "Displays information about the notebook.",
    },
    {
        "command": "add-note",
        "usage": "add-note",
        "exmp": "add-note",
        "desc": "Adds a new note with the specified name and text.",
    },
    {
        "command": "edit-note",
        "usage": "edit-note <name>/<text>",
        "exmp": "edit-note name OR text",
        "desc": "Edits note name or text and save it in the Notebook",
    },
    {
        "command": "delete-note",
        "usage": "delete-note <name>/<all>",
        "exmp": "delete-note test OR all",
        "desc": "Removes a specific note by name or all notes if name is 'all'",
    },
    {
        "command": "search-note",
        "usage": "search-note <name>",
        "exmp": "search-note <name>",
        "desc": "Searches for a note by the note name or keyword",
    },
    {
        "command": "all-notes",
        "usage": "all notes",
        "exmp": "all-notes",
        "desc": "Displays all notes in the notebook.",
    },
    {
        "command": "close",
        "usage": "close",
        "exmp": "close",
        "desc": "Saving Addressbook data and closes the application.",
    },
    {
        "command": "exit",
        "usage": "exit",
        "exmp": "exit",
        "desc": "Exits the address book application.",
    },
    {
        "command": "add-tag",
        "usage": "add-tag <note-name> <tag>",
        "exmp": "add-tag todo @task",
        "desc": "Adds a single tag to a note. Use this command to associate a new tag with a note.",
    },
    {
        "command": "add-tags",
        "usage": "add-tags <note-name> <tag>",
        "exmp": "add-tags todo @task @task @task ",
        "desc": "Adds multiple tags to a note. Use this command to associate multiple new tags with a note.",
    },
    {
        "command": "remove-tag",
        "usage": "remove-tag <note_name> <tag>",
        "exmp": "remove-tag todo @task",
        "desc": "Removes a single tag from a note. Use this command to disassociate a tag from a note.",
    },
    {
        "command": "edit-tag",
        "usage": "edit-tag <note-name> <old_tag> <new_tag>",
        "exmp": "edit-tag todo @task @list",
        "desc": "Edits a tag on a note. Use this command to rename a tag associated with a note.",
    },
    {
        "command": "note-tags",
        "usage": "note-tags <note-name> ",
        "exmp": "note-tags todo",
        "desc": "Displays all tags associated with a note. Use this command to view the tags associated with a note.",
    },
    {
        "command": "all-tags",
        "usage": "all-tags",
        "exmp": "all-tags",
        "desc": "Displays all tags in the notebook. Use this command to view all available tags.",
    },
]

NOTEBOOK_INFO_TABLE_HEADERS = ["COMMAND", "USAGE", "EXAMPLE", "DESCRIPTION"]
NOTEBOOK_INFO_TABLE_DATA = [
    [c["command"], c["usage"], c["exmp"], c["desc"]] for c in NOTEBOOK_COMMANDS
]



import pyfiglet
import time
import sys

team_name = "NO NAME BOT"
bye_msg = "SEE YOU LATER"


def display_ascii_welcome_art(msg: str):
    """Display a welcome message in ASCII art.

    Args:
        msg (str) - The message to display.
    """
    ascii_art = pyfiglet.figlet_format(msg, font="big")
    print(ascii_art)


def live_print(text, delay=0.05):
    """Print text to the console with a delay.

    Args:
        text (str) - The text to print.
        delay (float) - The delay between each character in seconds.
    """
    for char in text:
        sys.stdout.write(char)
        sys.stdout.flush()
        time.sleep(delay)
    print()


def welcome():
    """Display a welcome message and ASCII art."""
    display_ascii_welcome_art(team_name)

    live_print(f"Welcome to the {team_name}.\n")
    live_print("Preparing your environment...\n")
    live_print(f"Initializing {team_name}...\n")
    time.sleep(0.2)
    live_print("Loading your Contact Book...\n")
    time.sleep(0.4)
    live_print("Loading your NoteBook...\n")
    time.sleep(0.3)
    live_print("Setup complete!\n")


def good_bye():
    """Display a good bye message and ASCII art."""
    live_print("Saving your Contact Book...\n")
    time.sleep(0.2)
    live_print("Saving your NoteBook...\n")
    time.sleep(0.3)

    display_ascii_welcome_art(bye_msg)
    live_print("Goodbye!\n")


if __name__ == "__main__":
    welcome()
    good_bye()



"""Decorators that handle input errors are collected here."""


def input_error(func):
    """Handles a missing arguments error

    Args:
        func (callable): function
    """

    def inner(*args, **kwargs):
        # parse_input_message = ("Invalid command")
        add_contact_message = (
            "Arguments are required. Print 'add name 1234567890', "
            "where name is contact's name, and 1234567890 is contacts phone number: "
            "10 digits numbers only."
        )
        change_contact_message = "Arguments are required. Print 'change name \
number new-number', where name is contact's name, and number is old \
number and then a new number: 10 digits numbers only."
        show_phone_message = "Argument is required. Print 'phone name', \
where name is contact's name."
        delete_contact_message = (
            "Argument is required. Print 'delete all', or 'delete name'."
        )
        add_birthday_message = "DD.MM.YYYY is format for birthday date."
        show_birthday_message = "Arguments are required. Print 'show-birthday name', \
where name is contact's name."
        show_add_phone_message = (
            "The phone number must contain 10 digits, only numbers are required"
        )
        show_edit_contact_input_message = (
            "Arguments are required. Enter edit-contact <name>"
        )
        search_contact_message = (
            "Arguments are required. Enter search-contact <name> | <email> | <phone>"
        )
        common_message = "Arguments are required."

        try:
            return func(*args, **kwargs)
        except IndexError as i:
            if func.__name__ == "show_phone":
                return show_phone_message
            if func.__name__ == "delete_contact":
                return delete_contact_message
            if func.__name__ == "show_birthday":
                return show_birthday_message
            if func.__name__ == "show_upcoming_birthdays":
                return str(i).strip("'")
            return common_message
        except ValueError:
            if func.__name__ == "add_contact":
                return add_contact_message
            if func.__name__ == "change_contact":
                return change_contact_message
            if func.__name__ == "show_phone":
                return show_phone_message
            if func.__name__ == "delete_contact":
                return delete_contact_message
            if func.__name__ == "add_birthday":
                return add_birthday_message
            if func.__name__ == "show_birthday":
                return show_birthday_message
            if func.__name__ == "add_phone_to_contact":
                return show_add_phone_message
            if func.__name__ == "edit_contact_input":
                print(show_edit_contact_input_message)
            if func.__name__ == "search_contact":
                return search_contact_message
            return common_message
        except KeyError as e:
            if func.__name__ == "edit_contact_input":
                print(str(e).strip('"'))

            return str(e).strip("'")

    return inner


def empty_contact_list(func):
    """handles an empty contact list error

    Args:
        func (callable): any function that depends on the fullness of the contact list
    """

    def inner(*args, **kwargs):
        if len(args) <= 1:
            if len(args[0]) == 0:
                return "The contacts list is empty. \
Type 'add-contact' to add your first contact."
        elif len(args) > 1:
            if len(args[1]) == 0:
                return "The contacts list is empty. \
Type 'add-contact' to add your first contact."
        return func(*args, **kwargs)

    return inner



"""functions for work with pickle"""

import logging
import pickle
from pathlib import Path

from classes import AddressBook


def save_data(
    book: AddressBook, filename: str = "addressbook.pkl", log_msg=True
) -> None:
    """
    Save data to a pickle file.

    Args:
        book (AddressBook): The object to be saved (e.g., an instance of AddressBook).
        filename (str): The name of the file to save the data to.
        Default is "addressbook.pkl".

    Raises:
        Exception: If an error occurs during the saving process, it will be logged.
    """
    try:
        with Path(filename).open("wb") as f:
            pickle.dump(book, f)
        if log_msg:
            logging.info("Data successfully saved to %s", filename)
    except Exception as e:
        logging.error("Error occurred while saving data to %s: %s", filename, e)


def load_data(filename: str = "addressbook.pkl") -> AddressBook:
    """
    Load data from a pickle file.

    Args:
        filename (str): The name of the file to load the data from.
        Default is "addressbook.pkl".

    Returns:
        Any: The loaded data, or a new instance of AddressBook
        if the file is not found or an error occurs.
    """
    file_path = Path(filename)
    if not file_path.exists():
        logging.warning(
            "File %s not found. Returning a new AddressBook instance.", filename
        )
        return AddressBook()

    try:
        with file_path.open("rb") as f:
            data = pickle.load(f)
        logging.info("Data successfully loaded from %s", filename)
        return data
    except (pickle.UnpicklingError, EOFError) as e:
        logging.error("Error occurred while loading data from %s: %s", filename, e)
        return AddressBook()
    except Exception as e:
        logging.critical(
            "Unknown error occurred while loading data from %s: %s", filename, e
        )
        return AddressBook()



"""A common function that shows possible bot commands in table view."""

import shutil
from tabulate import tabulate


def table_show(headers, data, centred=False):
    """Show data in table view."""
    table = tabulate(
        data,
        headers,
        tablefmt="mixed_grid",
        stralign="left",
    )
    if centred:
        terminal_width = shutil.get_terminal_size().columns
        centered_table = ""
        for line in table.splitlines():
            padding = (terminal_width - len(line)) // 2
            centered_table += " " * max(0, padding) + line + "\n"

        return centered_table
    else:
        return table



"""Record class"""

from datetime import datetime
from classes.birthday import Birthday
from classes.name import Name
from classes.phone import Phone
from classes.email import Email
from classes.address import Address


class Record:
    """
    A class for storing contact information, including name and phone numbers.

    Attributes:
        * name (Name) - The contact's name.
        * phones (list of Phone) - A list of the contact's phone numbers.
    """

    def __init__(self, name):
        self.name = Name(name)
        self.phones = []
        self.birthday = None
        self.email = None
        self.address = None

    def __str__(self):
        contact_string = f"Contact name: {self.name.value}"
        if len(self.phones) > 0:
            contact_string += f", phones: {', '.join(p.value for p in self.phones)}"

        if self.birthday:
            contact_string += f", birthday: {self.birthday}"

        if self.email:
            contact_string += f", email: {self.email}"

        if self.address:
            contact_string += f", address: {self.address}"

        return contact_string

    def add_phone(self, number: str):
        """
        Add a phone number to the record.

        Args:
            * phone (str) - The phone number to be added.
        """
        self.phones.append(Phone(number))

    def remove_phone(self, number: str):
        """
        Remove a phone number from the record.

        Args:
            * phone (str) - The phone number to be removed.
        """

        for phone in self.phones:
            if str(phone.value) == number:
                self.phones.remove(phone)

    def edit_phone(self, old_number: str, new_number: str):
        """
        Edit a phone number in the record.

        Args:
            * old_phone (str) - The phone number to be replaced.
            * new_phone (str) - The new phone number to replace the old one.
        """

        self.phones = list(
            map(
                lambda phone: Phone(new_number) if phone.value == old_number else phone,
                self.phones,
            )
        )

    def find_phone(self, number):
        """
        Find a phone number in the record.

        Args:
            * phone (str) - The phone number to find.

        Returns:
            * Phone - The phone object if found, None otherwise.
        """

        for phone in self.phones:
            if phone.value == number:
                return phone

    def add_birthday(self, date_of_birthday):
        """Add a birthday to the record."""
        self.birthday = Birthday(date_of_birthday)

    def add_email(self, email):
        """Add an email to the record."""
        self.email = Email(email)

    def add_address(self, address_string):
        """Add an address to the record."""
        self.address = Address(address_string)

    def edit_name(self, new_name):
        """Edit name in the record."""
        self.name.value = new_name

    def edit_email(self, new_email):
        """Edit email in the record."""
        if self.email is None:
            self.email = Email(new_email)
        self.email.value = new_email

    def edit_address(self, new_address):
        """Edit address in the record."""
        if self.address is None:
            self.address = Address(new_address)
        self.address.value = new_address

    def edit_birthday(self, new_birthday):
        """Edit birthday in the record."""
        if self.birthday is None:
            self.add_birthday(new_birthday)
        self.birthday.value = datetime.strptime(new_birthday, "%d.%m.%Y")

    def remove_email(self):
        """Delete email in the record."""
        self.email = None

    def remove_address(self):
        """Delete address in the record."""
        self.address = None

    def remove_birthday(self):
        """Delete birthday in the record."""
        self.birthday = None




class Tag:
    """ Tag class"""

    """ Constructor """
    def __init__(self, name: str):
        self.__name = name

    """ To str """
    def __str__(self):
        return f"Tag: {self.__name}"
    
    """ Getter for name"""
    @property
    def name(self):
        return self.__name
    
    """ Setter for name"""
    @name.setter
    def name(self, new_name: str):
        self.__name = new_name


from setuptools import setup, find_packages

setup(
    name='no_name_bot',
    version='1.6',
    description="Python noname bot",
    url='https://github.com/KopchukKateryna/project-S8cceS7f',
    author='Kateryna Kopchuk, Ivan Zarichnyy, Natalya Karpova, Zakhar Veremchuk, Oleg Sokolenko',
    license='MIT',
    packages=['handlers', 'helpers', 'constants','classes'],
    install_requires=[
        'autopep8==2.3.1',
        'black==24.8.0',
        'cfgv==3.4.0',
        'click==8.1.7',
        'colorlog==6.8.2',
        'distlib==0.3.8',
        'filelock==3.15.4',
        'flake8==7.1.1',
        'identify==2.6.0',
        'importlib-metadata==8.2.0',
        'mccabe==0.7.0',
        'mypy-extensions==1.0.0',
        'nodeenv==1.9.1',
        'packaging==24.1',
        'pathspec==0.12.1',
        'platformdirs==4.2.2',
        'pre-commit==3.8.0',
        'prompt-toolkit==3.0.47',
        'pycodestyle==2.12.1',
        'pyfiglet==1.0.2',
        'pyflakes==3.2.0',
        'pyreadline==2.1',
        'pyreadline3==3.4.1',
        'PyYAML==6.0.2',
        'tabulate==0.9.0',
        'tomli==2.0.1',
        'typing-extensions==4.12.2',
        'virtualenv==20.26.3',
        'wcwidth==0.2.13',
        'yapf==0.40.2',
        'zipp==3.20.0',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ],
    entry_points={
        'console_scripts': [
            'no_name_bot = main:main',
        ],
    },
)


"""class AddressBook"""

from collections import UserDict
from datetime import datetime, timedelta


class AddressBook(UserDict):
    """
    Stores the contact book and methods for working with it.
    Inherits from UserDict.
    """

    def __str__(self):
        lines = [str(record) for record in self.data.values()]
        return "\n".join(lines)

    def add_record(self, record):
        """
        Add a contact record to the address book.

        Args:
            * record (Record) - The contact record to be added.
        """
        if record.name.value in self.data:
            raise KeyError(f"Record with name '{record.name.value}' already exists.")
        self.data[record.name.value] = record

    def find(self, name):
        """
        Finds a record by name.

        Args:
            * name(str) - The name of the contact to find.

        Returns:
            * Record - The contact record if found
        """
        record = self.data.get(name, None)
        return record

    def find_email(self, email):
        """
        Finds a record by email.

        Args:
            * email(str) - The email of the contact to find.

        Returns:
            * Record - The contact record if found
        """
        return next(
            (
                record
                for record in self.data.values()
                if record.email and record.email.value == email
            ),
            None,
        )

    def delete(self, name):
        """
        Deletes a record by name.

        Args:
            * name(str) - The name of the record to be deleted.
        """
        if name not in self.data:
            raise KeyError(f"Record with name '{name}' not found.")
        del self.data[name]

    def get_upcoming_birthdays(self, days):
        """
        Creates a list of contacts to wish happy birthday within the next 'days' days.

        Args:
            days (int): The number of days from today to check for upcoming birthdays.

        Returns:
            list: List of dicts containing contacts with upcoming birthdays.
        """
        today_datetime = datetime.today().date()
        upcoming_birthdays = []

        for name, record in self.data.items():
            if record.birthday:
                user_birthday = datetime.strptime(
                    str(record.birthday), "%d.%m.%Y"
                ).date()
                birthday_prepared = user_birthday.replace(year=today_datetime.year)
                if birthday_prepared < today_datetime:
                    birthday_prepared = birthday_prepared.replace(
                        year=today_datetime.year + 1
                    )

                if (birthday_prepared - today_datetime).days > days:
                    continue

                if birthday_prepared.weekday() > 5:
                    birthday_prepared += timedelta(days=7 - birthday_prepared.weekday())

                congratulation_date_str = birthday_prepared.strftime("%d.%m.%Y")
                upcoming_birthdays.append(
                    {"name": name, "congratulation_date": congratulation_date_str}
                )

        return upcoming_birthdays

    def find_by_phone(self, phone):
        """
        Finds a record by phone number.
        Args:
        * phone(str) - The phone number to search for.
        Returns:
        * dict - The record with the matching phone number.
        """
        return next(
            (
                record
                for record in self.data.values()
                for p in record.phones
                if p.value == phone
            ),
            None,
        )
